/*
 * Punix, Puny Unix kernel
 * Copyright 2005 Chris Williams
 * 
 * $Id: sys_file.c,v 1.3 2005/03/28 21:39:33 chris Exp $
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*-
 * Copyright (c) 1986, 1989, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>

#include <sys/proc.h>
#include <sys/file.h>
#include <sys/punix.h>
#include <sys/process.h>

STARTUP(void sys_getdtablesize())
{
	P.p_retval = NR_OPEN;
}

/* FIXME: read and write are just quick hacks to produce results */
STARTUP(void sys_read())
{
	struct a {
		int	fd;
		void	*buf;
		size_t	count;
	} *ap = (struct a *)P.p_arg;
	
	if (ap->fd != 0)
		P.p_error = EBADF;
}

STARTUP(void sys_write())
{
	struct a {
		int	fd;
		void	*buf;
		size_t	count;
	} *ap = (struct a *)P.p_arg;
	
	int kputchar(int c);
	char *buf;
	ssize_t count;
	if (ap->fd != 1 && ap->fd != 2) {
		P.p_error = EBADF;
		return;
	}
	
	/* we should check that ap->count doesn't overflow count */
	count = ap->count;
	P.p_retval = count;
	
	for (buf = ap->buf; count; --count, ++buf)
		kputchar(*buf);
}

STARTUP(void sys_open())
{
	struct a {
		const char *pathname;
		int flags;
		mode_t mode;
	} *ap = (struct a *)P.p_arg;
	
	/* FIXME: write this */
	(void)ap;
	P.p_error = EMFILE;
}

STARTUP(static int close(struct file *fp))
{
	int error;
	
	if (!fp)
		return 0;
	if (fp->f_count > 1) {
		--fp->f_count;
		return 0;
	}
	/* FIXME: finish writing this */
	error = 0 /* Fops[fp->f_type]->fo_close(fp) */;
	return error;
}

STARTUP(void sys_close())
{
	struct a {
		int fd;
	} *ap = (struct a *)P.p_arg;
	struct file *fp;
	
	GETF(fp, ap->fd);
	P.p_ofile[ap->fd] = NULL;
	while (P.p_lastfile >= 0 && P.p_ofile[P.p_lastfile] == NULL)
		--P.p_lastfile;
	
	P.p_error = close(fp);
}

/* allocate a process file descriptor */
STARTUP(static int pfalloc(int i))
{
	/* look for lowest free fd */
	for (; i < NR_OPEN; ++i)
		if (!P.p_ofile[i]) {
			P.p_retval = i;
			P.p_ofileflags[i] = 0;
			if (i > P.p_lastfile)
				P.p_lastfile = i;
			return i;
		}
	P.p_error = EMFILE;
	return -1;
}

#if 0
static struct file *falloc()
{
	struct file *fp;
	int fd;
	
	if ((fd = pfalloc(0)) < 0)
		return NULL;
	/* FIXME: finish this */
}
#endif

/*
static struct file *getf(f)
{
	struct file *fp;
	
	if ((unsigned)f >= NR_OPEN || (fp = P.p_ofile[f])) {
		P.p_error = EBADF;
		return NULL;
	}
	
	return fp;
}
*/
	
STARTUP(static void dupit(register int fd, register struct file *fp, int flags))
{
	P.p_ofile[fd] = fp;
	P.p_ofileflags[fd] = flags;
	++fp->f_count;
	if (fd > P.p_lastfile)
		P.p_lastfile = fd;
}

STARTUP(void sys_dup())
{
	struct a {
		int	oldfd;
	} *ap = (struct a *)P.p_arg;
	int newfd;
	struct file *oldfp;
	
	GETF(oldfp, ap->oldfd);
	
	if ((newfd = pfalloc(0)) < 0)
		return;
	
	dupit(newfd, oldfp, P.p_ofileflags[ap->oldfd] & ~FD_CLOEXEC);
}

STARTUP(void sys_dup2())
{
	struct a {
		int	oldfd;
		int	newfd;
	} *ap = (struct a *)P.p_arg;
	struct file *oldfp;
	
	GETF(oldfp, ap->oldfd);
	
	if ((unsigned)ap->newfd >= NR_OPEN) {
		P.p_error = EBADF;
		return;
	}
	
	if (P.p_ofile[ap->newfd])
		close(P.p_ofile[ap->newfd]);
	
	dupit(ap->newfd, oldfp, P.p_ofileflags[ap->newfd] & ~FD_CLOEXEC);
	P.p_retval = ap->newfd;
}

STARTUP(void sys_lseek())
{
	struct a {
		int fd;
		off_t offset;
		int whence;
	} *ap = (struct a *)P.p_arg;
	off_t pos;
	struct file *fp;
	
	GETF(fp, ap->fd);
	
	if (fp->f_type == DTYPE_PIPE) {
		P.p_error = ESPIPE;
		return;
	}
	
	pos = P.p_ofile[ap->fd]->f_offset;
	switch (ap->whence) {
		case SEEK_SET:
			pos = ap->offset;
			break;
		case SEEK_CUR:
			pos += ap->offset;
			break;
		case SEEK_END:
			/* FIXME: find end first */
			pos += ap->offset;
			break;
		default:
			P.p_error = EINVAL;
			return;
	}
	
	P.p_ofile[ap->fd]->f_offset = pos;
	P.p_retval = (long)pos;
}

#if 0
/* This is from Uzix */
/* Getinode() returns the inode pointer associated with a user's
 * file descriptor, checking for valid data structures
 */
inoptr getinode(int fd)
{
	inoptr inoindex;
	struct file *fp; //uchar oftindex = UDATA(u_files)[fd];
	
	GETF(fp, fd);
	/*if (oftindex >= OFTSIZE) // check that fp is within bounds
		panic(gtinobadoft, oftindex, " for", fd);*/
	if ((inoindex = of_tab[oftindex].o_inode) < i_tab ||
	    inoindex >= i_tab + ITABSIZE)
		panic(gtinobadoft, oftindex, ", inode", inoindex);
	magic(inoindex);
	return inoindex;
}
#endif

#if 0
void fstat()
{
	struct a {
		int fd;
		struct stat	*buf;
	} *ap = (struct a *)P.p_arg;
	struct stat buf;
	
	GETF(fp, ap->fd);
	switch (fp->f_type) {
		case DTYPE_PIPE:
		case DTYPE_INODE:
			P.p_error = ino_stat((struct inode *)fp->f_data, &buf);
			if (fp->f_type == DTYPE_PIPE)
				buf.st_size -= fp->f_offset;
			break;
		
		default:
			P.p_error = EINVAL;
			break;
	}
	if (!P.p_error)
		P.p_error = copyout(ap->buf, &buf, sizeof(buf));
}

static int fioctl(struct file *fp, unsigned cmd, char *value)
{
	return Fops[fp->f_type]->fo_ioctl(fp, cmd, value);
}

static int fgetown(fp, valuep)
	struct file *fp;
	int *valuep;
{
	int error;
	
	error = fioctl(fp, (unsigned)TIOCGPGRP, (char *)valuep);
	*valuep = -*valuep;
	return error;
}

static int fsetown(fp, value)
	struct file *fp;
	int value;
{
	if (value > 0) {
		struct proc *p = pfind(value);
		if (!p)
			return ESRCH;
		value = p->p_pgrp;
	} else
		value = -value;
	return fioctl(fp, (unsigned)TIOCSPGRP, (char *)&value);
}

void fcntl()
{
	struct a {
		int	fd;
		int	cmd;
		union {
			long	arg;
			struct flock	*lock;
		}
	} *ap = (struct a *)P.p_arg;
	struct file *fp;
	long i;
	char *ofp;
	
	GETF(fp, ap->fd);
	ofp = &P.p_ofileflags[ap->fd];
	switch (ap->cmd) {
		case F_DUPFD:
			i = ap->arg;
			if (i < 0 || i >= NR_OPEN) {
				P.p_error = EINVAL;
				return;
			}
			if ((i = pfalloc(i)) < 0)
				return;
			dupit(i, fp, *ofp & ~FD_CLOEXEC);
			break;
		case F_GETFD:
			P.p_retval = *ofp & FD_CLOEXEC;
			break;
		case F_SETFD:
			*pop = (*ofp & ~FD_CLOEXEC) | (ap->arg & FD_CLOEXEC);
			break;
		case F_GETFL:
			P.p_retval = OFLAGS(fp->f_flag);
			break;
		case F_SETFL:
			fp->f_flag &= ~FCNTLFLAGS;
			fp->f_flag |= (FFLAGS(ap->arg)) & FCNTLFLAGS;
			if ((P.p_error = fset(fp, FNONBLOCK, fp->f_flag & NONBLOCK)))
				break;
			if ((P.p_error = fset(fp, O_ASYNC, fp->f_flag & O_ASYNC)))
				fset(fp, FNONBLOCK, 0);
			break;
		case F_GETOWN:
			P.p_error = fgetown(fp, &P.p_retval);
			break;
		case F_SETOWN:
			P.p_error = fsetown(fp, ap->arg);
			break;
		default:
			P.p_error = EINVAL;
	}
}
#endif

#if 0
/*
 * UZIX - UNIX Implementation for MSX
 * (c) 1997-2001 Arcady Schekochikhin
 *		 Adriano C. R. da Cunha
 *
 * UZIX is based on UZI (UNIX Zilog Implementation)
 * UZI is a UNIX kernel clone written for Z-80 systems.
 * All code is public domain, not being based on any AT&T code.
 *
 * The author, Douglas Braun, can be reached at:
 *	7696 West Zayante Rd.
 *	Felton, CA 95018
 *	oliveb!intelca!mipos3!cadev4!dbraun
 *
 * This program is under GNU GPL, read COPYING for details
 *
 */

/********************************************
 SYSCALL pipe(int filedes[2]);
*******************************************/

static void filldesc(uchar oft, uchar access, inoptr ino) {
	oft_t *op;
	
	op = &of_tab[oft];
	op->o_ptr = 0;
	op->o_inode = ino;
	op->o_access = access;
}

long dopipe(int filedes[2]) {
	uchar u1, u2, oft1 = (uchar)-1, oft2 = (uchar)-1;
	inoptr ino;

	/* uf_alloc doesn't alloc an entry in user file table,
	   just returns the first free entry. so you must set it
	   before allocating another, or you'll get the same entry
	*/

	if ((ino = i_open(root_dev, NULLINO)) == NULL) {
		UERR = ENOMEM;
		goto Err;
	}
	if ((u1 = uf_alloc()) == (uchar)-1 ||
	    (oft1 = oft_alloc()) == (uchar)-1) {
		UERR = ENOENT;
		goto Err1;
	}
	current->p_files[u1] = oft1;	/* read side */
	filldesc(oft1, O_RDONLY, ino);

	if ((u2 = uf_alloc()) == (uchar)-1 ||
	    (oft2 = oft_alloc()) == (uchar)-1) {
		current->p_files[u1] = -1; 	/* free user file table entry */
		UERR = ENOENT;
		goto Err2;
	}
	/* fill in pipe descriptors */
	current->p_files[u2] = oft2;	/* write side */
	filldesc(oft2, O_WRONLY, ino);

	/* No permissions necessary on pipes */
	ino->i_mode = S_IFPIPE | 0777;
	ino->i_nlink = 0;	/* a pipe is not in any directory */
	++ino->i_refs;			/* count two sides of pipe */

	filedes[0] = u1;
	filedes[1] = u2;

	return 0;

Err2:	if (oft2 != (uchar)-1) oft_deref(oft2);
Err1:	if (oft1 != (uchar)-1) oft_deref(oft1);
Err:	return (-1);
}
#endif
