/*
 * Punix, Puny Unix kernel
 * Copyright 2005 Chris Williams
 * 
 * $Id: sys_proc.c,v 1.4 2005/03/28 21:43:11 chris Exp $
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <signal.h>
#include <sys/proc.h>
#include <sys/punix.h>
#include <sys/process.h>

#include <errno.h>

struct execa {
	char *filename;
	char **argv;
	char **envp;
};

STARTUP(void sys_execve())
{
	/* close all files marked "close on exec" */
	/* reset all signals to default (unless it's ignored) */
	/* setup environment */
	/* setup arguments */
	/* free process's memory */
	/* unblock parent if blocked by us with vfork */
	P.p_error = ENOMEM;
	/* FIXME: write this */
}

STARTUP(void sys_execv())
{
	((struct execa *)P.p_arg)->envp = NULL;
	sys_execve();
}

STARTUP(void sys_exit())
{
	struct a {
		int status;
	} *ap = (struct a *)P.p_arg;
	
	CURRENT->p_xstat = ap->status;
	/* clean up resources used by proc */
	/*
	 * we probably don't need to free memory allocations, as we probably
	 * don't need to increment ref counts to mem allocs upon vfork
	 * This is because a vfork parent is blocked until the child exits or
	 * execs anyway
	 */
	/* close all fd's */
	/* any more we have to do? */
	/* FIXME: write this */
	
	/* FIXME: add signals to appropriate global header file and fix this */
	sendsig(CURRENT->p_pptr, 42/*SIGCHLD*/);
	
	++CURRENT->p_insys;
	CURRENT->p_status = P_ZOMBIE;
	/* if parent is blocked by us with vfork, we need to wake it up */
	nextproc();
	panic("exit: nextproc failed");
}

STARTUP(void sys_fork())
{
	P.p_error = ENOMEM;	/* we lie (snicker) */
}

STARTUP(void sys_vfork())
{
	P.p_error = ENOMEM;
}

/* should I generalise the next two routines? */
/* for example:
void *nalloc(void **pool)
{
	struct {
		struct *np next;
	} np **f = (struct np **)pool;
	if (*f)
		*f = (*f)->next;
	
	return (void *)*f;
}

void nfree(void **pool, void *n)
{
	struct {
		struct *np next;
	} np **f = (struct np **)pool,
	  *p = (struct np *)n;
	
	if (p) {
		p->next = *f;
		*f = p;
	}
}

struct proc *palloc()
{ return nalloc(&GLOBAL_freeproc); }

void pfree(struct proc *p)
{ return nfree(&GLOBAL_freeproc, p); }
*/

/* allocate a process structure */
STARTUP(struct proc *palloc())
{
	struct proc *p = GLOBAL_freeproc;
	if (p)
		GLOBAL_freeproc = p->p_next;
	
	return p;
}

/* free a process structure */
STARTUP(void pfree(struct proc *p))
{
	if (p) {
		p->p_next = GLOBAL_freeproc;
		GLOBAL_freeproc = p;
	}
}

/* find an unused process id */
STARTUP(int pidget())
{
	return -1;
}

/*
 * newproc()
 * in-kernel fork function
 * returns 0 for old proc, 1 for new proc
 *
 */
#if 1
STARTUP(int newproc())
{
	P.p_pid = 1;
	return 1; /* FIXME: this is just a stub for kmain() */
}
#else
STARTUP(int newproc())
{
	struct proc *p, *cp;
	int pid;
	
	pid = pidget();
	if ((p = palloc()) == NULL)
		panic("no procs");
	
	cp = &P;
	
	/* are these all I need? */
	p->p_uid = cp->p_uid;
	p->p_pgrp = cp->p_pgrp;
	p->p_nice = cp->p_nice;
	p->p_pid = pid;
	p->p_ppid = cp->p_pid;
	p->p_pptr = cp;
	p->p_time = 0;
	p->p_cpu = 0;
	p->p_sigmask = cp->p_sigmask;
	p->p_sigcatch = cp->p_sigcatch;
	p->p_sigignore = cp->p_sigignore;
	
	p->p_next = GLOBAL_allproc;
	GLOBAL_allproc = p;
	
	/* increment reference counts on all shared objects */
	
	/* */
	/* FIXME: finish writing this */
	return 0;
}
#endif
